/*
 * Compressed RAM based swap device
 *
 * Copyright (C) 2008, 2009  Nitin Gupta
 *
 * This code is released using a dual license strategy: BSD/GPL
 * You can choose the licence that better fits your requirements.
 *
 * Released under the terms of 3-clause BSD License
 * Released under the terms of GNU General Public License Version 2.0
 *
 * Project home: http://compcache.googlecode.com
 */

#ifndef _RAMZSWAP_H_
#define _RAMZSWAP_H_

#include "ramzswap_ioctl.h"
#include "sub-projects/allocators/xvmalloc-kmod/xvmalloc.h"


#define MAX_NUM_DEVICES 32


struct zobj_header {
#if 0
	u32 table_idx;
#endif
};




#define DEFAULT_DISKSIZE_PERC_RAM	25
#define DEFAULT_MEMLIMIT_PERC_RAM	15


#define MAX_CPAGE_SIZE_BDEV	(PAGE_SIZE / 2)


#define MAX_CPAGE_SIZE_NOBDEV	(PAGE_SIZE / 4 * 3)





#define SECTOR_SHIFT		9
#define SECTOR_SIZE		(1 << SECTOR_SHIFT)
#define SECTORS_PER_PAGE_SHIFT	(PAGE_SHIFT - SECTOR_SHIFT)
#define SECTORS_PER_PAGE	(1 << SECTORS_PER_PAGE_SHIFT)


#define C "ramzswap: "


#define NOP	do { } while (0)

#if defined(CONFIG_BLK_DEV_RAMZSWAP_STATS)
#define STATS
#endif

#if defined(STATS)
#define stat_inc(stat)			((stat)++)
#define stat_dec(stat)			((stat)--)
#define stat_inc_if_less(stat, val1, val2) \
				((stat) += ((val1) < (val2) ? 1 : 0))
#define stat_dec_if_less(stat, val1, val2) \
				((stat) -= ((val1) < (val2) ? 1 : 0))
#else	
#define stat_inc(x)			NOP
#define stat_dec(x)			NOP
#define stat_inc_if_less(x, v1, v2)	NOP
#define stat_dec_if_less(x, v1, v2)	NOP
#endif	


enum rzs_pageflags {
	
	RZS_UNCOMPRESSED,

	
	RZS_ZERO,

	__NR_RZS_PAGEFLAGS,
};




struct table {
	u32 pagenum;
	u16 offset;
	u8 count;	
	u8 flags;
};

struct ramzswap_backing_extent {
	u32 phy_pagenum;
	u32 num_pages;
};

struct ramzswap_stats {
	
	size_t compr_size;	
	
#if defined(STATS)
	u64 num_reads;		
	u64 num_writes;		
	u64 failed_reads;	
	u64 failed_writes;	
	u64 invalid_io;		
	u64 notify_free;	
	u64 pages_discard;	
	u32 pages_zero;		
	u32 pages_stored;	
	u32 good_compress;	
	u32 pages_expand;	
	u64 bdev_num_reads;	
	u64 bdev_num_writes;	
#endif
};

struct ramzswap {
	struct xv_pool *mem_pool;
	void *compress_workmem;
	void *compress_buffer;
	struct table *table;
	struct mutex lock;
	struct request_queue *queue;
	struct gendisk *disk;
	int init_done;
	int init_notify_callback;
	
	size_t memlimit;	
	
	size_t disksize;	

	struct ramzswap_stats stats;

	
	struct ramzswap_backing_extent *curr_extent;
	struct list_head backing_swap_extent_list;
	unsigned long num_extents;
	char backing_swap_name[MAX_SWAP_NAME_LEN];
	struct block_device *backing_swap;
	struct file *swap_file;
};


	
#endif
